package ru.dbotthepony.mc.otm.container

import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.network.NetworkEvent
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.menu.MatteryMenu
import ru.dbotthepony.mc.otm.set
import java.util.LinkedList
import java.util.function.Supplier

data class ItemFilterSlotPacket(val slotID: Int, val newValue: ItemStack) {
	fun write(buff: FriendlyByteBuf) {
		buff.writeInt(slotID)
		buff.writeItemStack(newValue, true)
	}

	fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true

		if (context.get().sender != null) {
			context.get().enqueueWork {
				playServer(context.get().sender!!)
			}
		} else {
			context.get().enqueueWork {
				playClient()
			}
		}
	}

	fun playClient() {
		val menu = minecraft.player?.containerMenu as? MatteryMenu ?: throw IllegalStateException("No MatteryMenu is open right now, can't handle ItemFilterSlotReplicaPacket")
		menu.filterSlots[slotID].set(newValue)
	}

	fun playServer(player: ServerPlayer) {
		val menu = player.containerMenu as? MatteryMenu ?: return LOGGER.error("No MatteryMenu is open right now, can't handle ItemFilterSlotReplicaPacket from $player")
		menu.filterSlots.getOrNull(slotID)?.set(newValue) //?: LOGGER.error("ItemFilterSlotReplicaPacket: unknown slot $slotID from $player in $menu!")
	}

	companion object {
		fun read(buff: FriendlyByteBuf): ItemFilterSlotPacket {
			return ItemFilterSlotPacket(buff.readInt(), buff.readItem())
		}

		private val LOGGER = LogManager.getLogger()
	}
}

data class ItemFilterNetworkSlot(val slotID: Int, val networkID: Int, val filter: ItemFilter?) {
	fun get() = filter?.get(slotID) ?: remote
	fun set(value: ItemStack) {
		if (filter != null)
			filter[slotID] = value
		else
			remote = value
	}

	private var remote: ItemStack = ItemStack.EMPTY

	fun sendChanges(full: Boolean = false): ItemFilterSlotPacket? {
		requireNotNull(filter) { "Invalid side" }

		if (full || !ItemStack.isSameItemSameTags(remote, get())) {
			remote = get()
			return ItemFilterSlotPacket(networkID, get())
		}

		return null
	}
}

class ItemFilter(
	val size: Int,
	private val modified: (slot: Int?, oldValue: ItemStack?, newValue: ItemStack?) -> Unit
) : INBTSerializable<CompoundTag> {
	private val filter = Array<ItemStack>(size) { ItemStack.EMPTY }
	private val linkedFilter = LinkedList<ItemStack>()

	var isWhitelist = false
		set(value) {
			if (value != field) {
				field = value
				modified.invoke(null, null, null)
			}
		}

	operator fun set(index: Int, value: ItemStack) {
		if (value.isEmpty && filter[index].isEmpty) {
			return
		}

		val old = filter[index]
		filter[index] = value.let { if (!it.isEmpty) it.copy().also { it.count = 1 } else it }

		if (!old.isEmpty)
			linkedFilter.remove(old)

		if (!filter[index].isEmpty)
			linkedFilter.add(filter[index])

		modified.invoke(index, old, filter[index])

	}

	operator fun get(index: Int): ItemStack = filter[index].copy()

	fun match(value: ItemStack): Boolean {
		if (value.isEmpty) {
			return false
		}

		if (linkedFilter.isEmpty()) {
			return !isWhitelist
		}

		if (isWhitelist) {
			for (item in linkedFilter) {
				if (item.`is`(value.item)) {
					return true
				}
			}

			return false
		}


		for (item in linkedFilter) {
			if (item.`is`(value.item)) {
				return false
			}
		}

		return true
	}

	override fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["items"] = ListTag().also {
				for (value in filter) {
					it.add(value.serializeNBT())
				}
			}

			it["is_whitelist"] = isWhitelist
		}
	}

	override fun deserializeNBT(nbt: CompoundTag?) {
		for (i in filter.indices)
			filter[i] = ItemStack.EMPTY

		if (nbt == null)
			return

		nbt.ifHas("items", ListTag::class.java) {
			for ((i, value) in it.withIndex()) {
				if (value is CompoundTag) {
					filter[i] = ItemStack.of(value)
				}
			}
		}

		isWhitelist = nbt.getBoolean("is_whitelist")
	}
}
